from jijmodeling.variables.variable import Variable
from jijmodeling.variables.binary import Binary
from jijmodeling.variables.dis_num import DisNum
from jijmodeling.variables.placeholder import Placeholder